import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens.plotters import plotter_util, grid_plotters, array_plotters


def plot_profile_image(
    plane,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_grid=False,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Profile Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="plane_profile_image",
):

    profile_image = plane.profile_image_from_grid(
        grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    )

    if plane.has_mass_profile:
        lines = plotter_util.get_critical_curve_and_caustic(
            obj=plane,
            grid=grid,
            plot_critical_curve=plot_critical_curves,
            plot_caustics=plot_caustics,
        )
    else:
        lines = None

    if not plot_grid:
        grid = None

    array_plotters.plot_array(
        array=profile_image,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        grid=grid,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_plane_image(
    plane,
    grid,
    plot_origin=True,
    positions=None,
    plot_grid=True,
    lines=None,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="plane_plane_image",
):

    plane_image = plane.plane_image_from_grid(grid=grid)

    if not plot_grid:
        grid = None

    if plot_origin:
        origin = plane_image.origin
    else:
        origin = None

    array_plotters.plot_array(
        array=plane_image,
        origin=origin,
        positions=positions,
        grid=grid,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_convergence(
    plane,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Convergence",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="plane_convergence",
):

    convergence = plane.convergence_from_grid(
        grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=plane,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=convergence,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_potential(
    plane,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Potential",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="plane_potential",
):

    potential = plane.potential_from_grid(
        grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=plane,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=potential,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_deflections_y(
    plane,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Deflections (y)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="plane_deflections_y",
):

    deflections = plane.deflections_from_grid(
        grid=grid, return_in_2d=False, return_binned=True
    )
    deflections_y = grid.mapping.scaled_array_2d_from_array_1d(
        array_1d=deflections[:, 0]
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=plane,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=deflections_y,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_deflections_x(
    plane,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Deflections (x)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="plane_deflections_x",
):

    deflections = plane.deflections_from_grid(
        grid=grid, return_in_2d=False, return_binned=True
    )
    deflections_x = grid.mapping.scaled_array_2d_from_array_1d(
        array_1d=deflections[:, 1]
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=plane,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=deflections_x,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_magnification(
    plane,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Plane Magnification",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="plane_magnification",
):

    magnification = plane.magnification_from_grid(
        grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=plane,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=magnification,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_image_and_source_plane_subplot(
    image_plane,
    source_plane,
    grid,
    points=None,
    plot_critical_curves=False,
    plot_caustics=False,
    axis_limits=None,
    units="arcsec",
    output_path=None,
    output_format="show",
    output_filename="image_and_source_plane_grids",
):

    rows, columns, figsize = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=2
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=image_plane, grid=grid, plot_critical_curve=True, plot_caustics=True
    )

    if plot_critical_curves:
        critical_curves = [lines[0]]
    else:
        critical_curves = None

    if plot_caustics:
        caustics = [lines[1]]
    else:
        caustics = None

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_plane_grid(
        plane=image_plane,
        grid=grid,
        axis_limits=axis_limits,
        points=points,
        lines=critical_curves,
        as_subplot=True,
        units=units,
        pointsize=3,
        xyticksize=16,
        titlesize=10,
        xlabelsize=10,
        ylabelsize=10,
        title="Image-plane Grid",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    source_plane_grid = image_plane.traced_grid_from_grid(grid=grid)

    plt.subplot(rows, columns, 2)

    plot_plane_grid(
        plane=source_plane,
        grid=source_plane_grid,
        axis_limits=axis_limits,
        points=points,
        lines=caustics,
        as_subplot=True,
        units=units,
        pointsize=3,
        xyticksize=16,
        titlesize=10,
        xlabelsize=10,
        ylabelsize=10,
        title="Source-plane Grid",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def plot_plane_grid(
    plane,
    grid,
    axis_limits=None,
    points=None,
    lines=None,
    as_subplot=False,
    units="arcsec",
    figsize=(12, 8),
    pointsize=3,
    title="Plane Grid",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="plane_grid",
):

    grid_plotters.plot_grid(
        grid=grid,
        points=points,
        axis_limits=axis_limits,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=plane.kpc_per_arcsec,
        figsize=figsize,
        pointsize=pointsize,
        xyticksize=xyticksize,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



from autolens.array import grids, mask as msk
from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld

from autolens.plotters import array_plotters

from test.simulation import simulation_util

import numpy as np

# Although we could test the deflection angles without using an image (e.al. by just making a grid), we have chosen to
# set this test up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

sub_size = 2
inner_radius_arcsec = 0.2
outer_radius_arcsec = 4.0

print("sub grid size = " + str(sub_size))
print("annular inner mask radius = " + str(inner_radius_arcsec) + "\n")
print("annular outer mask radius = " + str(outer_radius_arcsec) + "\n")

for data_resolution in ["HST_Up"]:

    print()

    imaging_data = simulation_util.load_test_imaging_data(
        data_type="lens_mass__source_smooth",
        data_resolution=data_resolution,
        psf_shape=(3, 3),
    )
    mask = al.Mask.circular_annular(
        shape=imaging_data.shape,
        pixel_scale=imaging_data.pixel_scale,
        inner_radius_arcsec=inner_radius_arcsec,
        outer_radius_arcsec=outer_radius_arcsec,
    )
    lens_data = al.LensData(imaging_data=imaging_data, mask=mask, sub_size=sub_size)

    print("Deflection angle run times for image type " + data_resolution + "\n")
    print("Number of points = " + str(lens_data.grid.shape[0]) + "\n")

    interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
        mask=lens_data.mask, grid=lens_data.grid, pixel_scale_interpolation_grid=0.05
    )

    print(
        "Number of interpolation points = "
        + str(interpolator.interp_grid.shape[0])
        + "\n"
    )

    ### EllipticalIsothermal ###

    mass_profile = al.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0
    )

    interp_deflections = mass_profile.deflections_from_grid(
        grid=interpolator.interp_grid
    )
    deflections = np.zeros((lens_data.grid.shape[0], 2))
    deflections[:, 0] = interpolator.interpolated_values_from_values(
        values=interp_deflections[:, 0]
    )
    deflections[:, 1] = interpolator.interpolated_values_from_values(
        values=interp_deflections[:, 1]
    )

    true_deflections = mass_profile.deflections_from_grid(grid=lens_data.grid)

    true_deflections_y_2d = lens_data.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
        sub_array_1d=true_deflections[:, 0]
    )
    true_deflections_x_2d = lens_data.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
        sub_array_1d=true_deflections[:, 1]
    )

    difference_y = deflections[:, 0] - true_deflections[:, 0]
    difference_x = deflections[:, 1] - true_deflections[:, 1]

    print("interpolation y error: ", np.mean(difference_y))
    print("interpolation y uncertainty: ", np.std(difference_y))
    print("interpolation y max error: ", np.max(difference_y))
    print("interpolation x error: ", np.mean(difference_x))
    print("interpolation x uncertainty: ", np.std(difference_x))
    print("interpolation x max error: ", np.max(difference_x))

    difference_y_2d = lens_data.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
        sub_array_1d=difference_y
    )
    difference_x_2d = lens_data.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
        sub_array_1d=difference_x
    )

    array_plotters.plot_array(array=true_deflections_y_2d)
    array_plotters.plot_array(array=difference_y_2d)

    array_plotters.plot_array(array=true_deflections_x_2d)
    array_plotters.plot_array(array=difference_x_2d)

    # difference_percent_y = (np.abs(difference_y) / np.abs(true_deflections[:,0]))*100.0
    # difference_percent_x = (np.abs(difference_x) / np.abs(true_deflections[:,1]))*100.0
    #
    # print("interpolation y mean percent difference: ", np.mean(difference_percent_y))
    # print("interpolation y std percent difference: ", np.std(difference_percent_y))
    # print("interpolation y max percent difference: ", np.max(difference_percent_y))
    # print("interpolation x mean percent difference: ", np.mean(difference_percent_x))
    # print("interpolation x std percent difference: ", np.std(difference_percent_x))
    # print("interpolation x mean percent difference: ", np.max(difference_percent_x))
    #
    # difference_percent_y_2d = lens_data.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
    #     sub_array_1d=difference_percent_y)
    # difference_percent_x_2d = lens_data.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
    #     sub_array_1d=difference_percent_x)
    #
    # array_plotters.plot_array(array=difference_percent_y_2d)
    # array_plotters.plot_array(array=difference_percent_x_2d)



from autolens.array import mask as msk
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.lens.lens_fit import lens_imaging_fit
from autolens.lens import lens_data as ld
from autolens.data.plotters import imaging_plotters
from autolens.lens.plotters import lens_imaging_fit_plotters
from test.simulation import simulation_util

imaging_data = simulation_util.load_test_imaging_data(
    data_type="lens_sis__source_smooth__offset_centre", data_resolution="LSST"
)


def fit_with_offset_centre(centre):

    mask = al.Mask.elliptical(
        shape=imaging_data.shape,
        pixel_scale=imaging_data.pixel_scale,
        major_axis_radius_arcsec=3.0,
        axis_ratio=1.0,
        phi=0.0,
        centre=centre,
    )

    # The lines of code below do everything we're used to, that is, setup an image and its al.ogrid, mask it, trace it
    # via a tracer, setup the rectangular mapper, etc.
    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
        ),
    )
    source_galaxy = al.Galaxy(
        redshift=1.0,
        pixelization=al.pixelizations.VoronoiMagnification(shape=(20, 20)),
        regularization=al.regularization.Constant(coefficient=1.0),
    )

    lens_data = al.LensData(imaging_data=imaging_data, mask=mask)

    pixelization_grid = source_galaxy.pixelization.traced_pixelization_grids_of_planes_from_grid(
        grid=lens_data.grid
    )

    grid_stack_with_pixelization_grid = lens_data.grid.new_grid_stack_with_grids_added(
        pixelization=pixelization_grid
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[lens_galaxy, source_galaxy],
        image_plane_grid=grid_stack_with_pixelization_grid,
    )
    fit = al.LensImageFit.from_lens_data_and_tracer(lens_data=lens_data, tracer=tracer)

    return fit


fit = fit_with_offset_centre(centre=(1.0, 1.0))

lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
    should_plot_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(1.05, 1.05))


lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
    should_plot_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(1.1, 1.1))

lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
    should_plot_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(0.95, 0.95))

lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
    should_plot_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(5.9, 5.9))

lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
    should_plot_image_plane_pix=True,
)



import time

from autolens.array import mask as msk
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing
from autolens.lens.lens_fit import lens_imaging_fit
from autolens.lens.util import lens_fit_util

from test.simulation import simulation_util

repeats = 10

print("Number of repeats = " + str(repeats))
print()

sub_size = 4
radius_arcsec = 3.0
psf_shape = (21, 21)

print("sub grid size = " + str(sub_size))
print("circular mask radius = " + str(radius_arcsec) + "\n")
print("psf shape = " + str(psf_shape) + "\n")

lens_galaxy = al.Galaxy(
    light=al.light_profiles.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.9,
        phi=45.0,
        intensity=0.5,
        effective_radius=0.8,
        sersic_index=4.0,
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
)

source_galaxy = al.Galaxy(
    light=al.light_profiles.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.4,
        effective_radius=0.5,
        sersic_index=1.0,
    )
)

for data_resolution in ["LSST", "Euclid", "HST", "HST_Up", "AO"]:

    imaging_data = simulation_util.load_test_imaging_data(
        data_type="lens_mass__source_smooth",
        data_resolution=data_resolution,
        psf_shape=psf_shape,
    )
    mask = al.Mask.circular(
        shape=imaging_data.shape,
        pixel_scale=imaging_data.pixel_scale,
        radius_arcsec=radius_arcsec,
    )
    lens_data = al.LensData(imaging_data=imaging_data, mask=mask, sub_size=sub_size)

    print("Light profile fit run times for image type " + data_resolution + "\n")
    print("Number of points = " + str(lens_data.grid.shape[0]) + "\n")

    start_overall = time.time()

    start = time.time()
    for i in range(repeats):
        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])
    diff = time.time() - start
    print("Time to Setup Tracer = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        blurred_profile_image_1d = lens_fit_util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=tracer.profile_image_from_grid,
            blurring_image_1d=tracer.profile_blurring_image,
            convolver=lens_data.convolver,
        )
        blurred_profile_image = lens_data.grid.mapping.scaled_array_2d_from_array_1d(
            array_1d=blurred_profile_image_1d
        )
    diff = time.time() - start
    print("Time to perform PSF convolution = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        al.LensDataFit(
            image_1d=lens_data.image_1d,
            noise_map_1d=lens_data.noise_map_1d,
            mask_1d=lens_data.mask_1d,
            model_image_1d=blurred_profile_image_1d,
        )
    diff = time.time() - start
    print("Time to perform fit (1D) = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        al.LensDataFit(
            image_1d=lens_data.unmasked_image,
            noise_map_1d=lens_data.unmasked_noise_map,
            mask_1d=lens_data.mask,
            model_image_1d=blurred_profile_image,
        )
    diff = time.time() - start
    print("Time to perform fit (2D) = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        al.LensImageFit.from_lens_data_and_tracer(lens_data=lens_data, tracer=tracer)
    diff = time.time() - start
    print("Time to perform complete fit = {}".format(diff / repeats))

    print()



import autolens as al
import os

import pytest


@pytest.fixture(name="imaging_plotter_path")
def make_imaging_plotter_setup():
    imaging_plotter_path = "{}/../../test_files/plotting/imaging/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    return imaging_plotter_path


def test__individual_attributes_are_output(
    imaging_data_7x7, positions_7x7, mask_7x7, imaging_plotter_path, plot_patch
):
    al.imaging_plotters.plot_image(
        imaging_data=imaging_data_7x7,
        positions=positions_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=imaging_plotter_path,
        output_format="png",
    )

    assert imaging_plotter_path + "imaging_image.png" in plot_patch.paths

    al.imaging_plotters.plot_noise_map(
        imaging_data=imaging_data_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=imaging_plotter_path,
        output_format="png",
    )

    assert imaging_plotter_path + "imaging_noise_map.png" in plot_patch.paths

    al.imaging_plotters.plot_psf(
        imaging_data=imaging_data_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=imaging_plotter_path,
        output_format="png",
    )

    assert imaging_plotter_path + "imaging_psf.png" in plot_patch.paths

    al.imaging_plotters.plot_signal_to_noise_map(
        imaging_data=imaging_data_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=imaging_plotter_path,
        output_format="png",
    )

    assert imaging_plotter_path + "imaging_signal_to_noise_map.png" in plot_patch.paths

    al.imaging_plotters.plot_imaging_subplot(
        imaging_data=imaging_data_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=imaging_plotter_path,
        output_format="png",
    )

    assert imaging_plotter_path + "imaging_data.png" in plot_patch.paths


def test__imaging_individuals__output_dependent_on_input(
    imaging_data_7x7, general_config, imaging_plotter_path, plot_patch
):
    al.imaging_plotters.plot_imaging_individual(
        imaging_data=imaging_data_7x7,
        should_plot_image=True,
        should_plot_psf=True,
        should_plot_absolute_signal_to_noise_map=True,
        output_path=imaging_plotter_path,
        output_format="png",
    )

    assert imaging_plotter_path + "imaging_image.png" in plot_patch.paths

    assert not imaging_plotter_path + "imaging_noise_map.png" in plot_patch.paths

    assert imaging_plotter_path + "imaging_psf.png" in plot_patch.paths

    assert (
        not imaging_plotter_path + "imaging_signal_to_noise_map.png" in plot_patch.paths
    )

    assert (
        imaging_plotter_path + "imaging_absolute_signal_to_noise_map.png"
        in plot_patch.paths
    )

    assert (
        not imaging_plotter_path + "imaging_potential_chi_squared_map.png"
        in plot_patch.paths
    )



import autofit as af
import autolens as al
from test.integration.tests import runner

test_type = "full_pipeline"
test_name = "hyper_with_lens_light_bg"
data_type = "lens_mass__source_smooth"
data_resolution = "LSST"


def make_pipeline(
    name,
    phase_folders,
    pipeline_pixelization=al.pixelizations.VoronoiBrightnessImage,
    pipeline_regularization=al.regularization.AdaptiveBrightness,
    optimizer_class=af.MultiNest,
):

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, light=al.light_profiles.EllipticalSersic)
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.3

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    class LensSubtractedPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light Sersic -> Sersic ##

            self.galaxies.lens.light = results.from_phase(
                "phase_1__lens_sersic"
            ).constant.galaxies.lens.light

            ## Lens Mass, Move centre priors to centre of lens light ###

            self.galaxies.lens.mass.centre = (
                results.from_phase("phase_1__lens_sersic")
                .variable_absolute(a=0.1)
                .galaxies.lens.light.centre
            )

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.hyper_image_sky = results.last.hyper_combined.constant.hyper_image_sky

            self.hyper_background_noise = (
                results.last.hyper_combined.constant.hyper_background_noise
            )

    phase2 = LensSubtractedPhase(
        phase_name="phase_2__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = False
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    phase2 = phase2.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    class LensSourcePhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light, Sersic -> Sersic ###

            self.galaxies.lens.light = results.from_phase(
                "phase_1__lens_sersic"
            ).variable.galaxies.lens.light

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens.mass = results.from_phase(
                "phase_2__lens_sie__source_sersic"
            ).variable.galaxies.lens.mass

            self.galaxies.lens.shear = results.from_phase(
                "phase_2__lens_sie__source_sersic"
            ).variable.galaxies.lens.shear

            ### Source Light, Sersic -> Sersic ###

            self.galaxies.source = results.from_phase(
                "phase_2__lens_sie__source_sersic"
            ).variable.galaxies.source

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.hyper_image_sky = results.last.hyper_combined.constant.hyper_image_sky

            self.hyper_background_noise = (
                results.last.hyper_combined.constant.hyper_background_noise
            )

    phase3 = LensSourcePhase(
        phase_name="phase_3__lens_sersic_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 75
    phase3.optimizer.sampling_efficiency = 0.3

    phase3 = phase3.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light & Mass, Sersic -> Sersic, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens.light = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).constant.galaxies.lens.light

            self.galaxies.lens.mass = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).constant.galaxies.lens.mass

            self.galaxies.lens.shear = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).constant.galaxies.lens.shear

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.hyper_image_sky = results.last.hyper_combined.constant.hyper_image_sky

            self.hyper_background_noise = (
                results.last.hyper_combined.constant.hyper_background_noise
            )

    phase4 = InversionPhase(
        phase_name="phase_4__initialize_magnification_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase4.optimizer.const_efficiency_mode = True
    phase4.optimizer.n_live_points = 20
    phase4.optimizer.sampling_efficiency = 0.8

    phase4 = phase4.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=False,
    )

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light & Mass, Sersic -> Sersic, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens

            ### Source Inversion, Inv -> Inv ###

            self.galaxies.source.pixelization = results.from_phase(
                "phase_4__initialize_magnification_inversion"
            ).constant.galaxies.source.pixelization

            self.galaxies.source.regularization = results.from_phase(
                "phase_4__initialize_magnification_inversion"
            ).constant.galaxies.source.regularization

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.hyper_image_sky = results.last.hyper_combined.constant.hyper_image_sky

            self.hyper_background_noise = (
                results.last.hyper_combined.constant.hyper_background_noise
            )

    phase5 = InversionPhase(
        phase_name="phase_5__lens_sersic_sie__source_magnification_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase5.optimizer.const_efficiency_mode = True
    phase5.optimizer.n_live_points = 75
    phase5.optimizer.sampling_efficiency = 0.2

    phase5 = phase5.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=False,
    )

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light & Mass, Sersic -> Sersic, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens.light = results.from_phase(
                "phase_5__lens_sersic_sie__source_magnification_inversion"
            ).constant.galaxies.lens.light

            self.galaxies.lens.mass = results.from_phase(
                "phase_5__lens_sersic_sie__source_magnification_inversion"
            ).constant.galaxies.lens.mass

            self.galaxies.lens.shear = results.from_phase(
                "phase_5__lens_sersic_sie__source_magnification_inversion"
            ).constant.galaxies.lens.shear

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.hyper_image_sky = results.last.hyper_combined.constant.hyper_image_sky

            self.hyper_background_noise = (
                results.last.hyper_combined.constant.hyper_background_noise
            )

    phase6 = InversionPhase(
        phase_name="phase_6_initialize_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=pipeline_pixelization,
                regularization=pipeline_regularization,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase6.optimizer.const_efficiency_mode = True
    phase6.optimizer.n_live_points = 20
    phase6.optimizer.sampling_efficiency = 0.8

    phase6 = phase6.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=True,
    )

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light & Mass, Sersic -> Sersic, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens = results.from_phase(
                "phase_5__lens_sersic_sie__source_magnification_inversion"
            ).variable.galaxies.lens

            ### Source Inversion, Inv -> Inv ###

            self.galaxies.source.pixelization = results.from_phase(
                "phase_6_initialize_inversion"
            ).hyper_combined.constant.galaxies.source.pixelization

            self.galaxies.source.regularization = results.from_phase(
                "phase_6_initialize_inversion"
            ).hyper_combined.constant.galaxies.source.regularization

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.galaxies.source.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
            )

            self.hyper_image_sky = results.last.hyper_combined.constant.hyper_image_sky

            self.hyper_background_noise = (
                results.last.hyper_combined.constant.hyper_background_noise
            )

    phase7 = InversionPhase(
        phase_name="phase_7__lens_sersic_sie__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=pipeline_pixelization,
                regularization=pipeline_regularization,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase7.optimizer.const_efficiency_mode = True
    phase7.optimizer.n_live_points = 75
    phase7.optimizer.sampling_efficiency = 0.2

    phase7 = phase7.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=True,
    )

    return al.PipelineImaging(
        name, phase1, phase2, phase3, phase4, phase5, phase6, phase7, hyper_mode=False
    )


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autolens as al
import numpy as np
import pytest


@pytest.fixture(name="lens_imaging_data_7x7")
def make_lens_imaging_data_7x7(imaging_data_7x7, sub_mask_7x7):
    return al.LensImagingData(imaging_data=imaging_data_7x7, mask=sub_mask_7x7)


@pytest.fixture(name="lens_imaging_data_6x6")
def make_lens_imaging_data_6x6(imaging_data_6x6, mask_6x6):
    return al.LensImagingData(imaging_data=imaging_data_6x6, mask=mask_6x6)


@pytest.fixture(name="lens_uv_plane_data_7")
def make_lens_uv_plane_data_7(uv_plane_data_7, sub_mask_7x7):
    return al.LensUVPlaneData(uv_plane_data=uv_plane_data_7, mask=sub_mask_7x7)


class TestAbstractLensData(object):
    def test__pixel_scale_interpolation_grid_input__grids_nclude_interpolators(
        self, sub_mask_7x7
    ):

        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, pixel_scale_interpolation_grid=1.0
        )

        grid = al.Grid.from_mask(mask=sub_mask_7x7)
        new_grid = grid.new_grid_with_interpolator(pixel_scale_interpolation_grid=1.0)

        assert (lens_imaging_data_7x7.grid == new_grid).all()
        assert (
            lens_imaging_data_7x7.grid.interpolator.vtx == new_grid.interpolator.vtx
        ).all()
        assert (
            lens_imaging_data_7x7.grid.interpolator.wts == new_grid.interpolator.wts
        ).all()

    def test__pixel_scale_binned_grid_is_input__correct_binned_up_grid_calculated(
        self, sub_mask_7x7, grid_7x7
    ):
        sub_mask_7x7.pixel_scale = 1.0
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, pixel_scale_binned_grid=1.0
        )

        assert lens_imaging_data_7x7.grid.binned.bin_up_factor == 1
        assert (
            lens_imaging_data_7x7.mask == lens_imaging_data_7x7.grid.binned.mask
        ).all()
        assert (lens_imaging_data_7x7.grid.binned == grid_7x7).all()
        assert (
            lens_imaging_data_7x7.grid.binned.binned_mask_1d_index_to_mask_1d_indexes
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        sub_mask_7x7.pixel_scale = 1.0
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, pixel_scale_binned_grid=1.9
        )

        assert lens_imaging_data_7x7.grid.binned.bin_up_factor == 1
        assert (
            lens_imaging_data_7x7.mask == lens_imaging_data_7x7.grid.binned.mask
        ).all()
        assert (
            lens_imaging_data_7x7.grid.binned.binned_mask_1d_index_to_mask_1d_indexes
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        sub_mask_7x7.pixel_scale = 1.0
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, pixel_scale_binned_grid=2.0
        )
        assert lens_imaging_data_7x7.grid.binned.bin_up_factor == 2
        assert (
            lens_imaging_data_7x7.grid.binned.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()
        assert (
            lens_imaging_data_7x7.grid.binned
            == np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])
        ).all()
        assert (
            lens_imaging_data_7x7.grid.binned.binned_mask_1d_index_to_mask_1d_indexes
            == np.array([[0, -1, -1, -1], [1, 2, -1, -1], [3, 6, -1, -1], [4, 5, 7, 8]])
        ).all()

        sub_mask_7x7.pixel_scale = 2.0
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, pixel_scale_binned_grid=1.0
        )

        assert lens_imaging_data_7x7.grid.binned.bin_up_factor == 1

        sub_mask_7x7.pixel_scale = 1.0
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, pixel_scale_binned_grid=None
        )

        assert lens_imaging_data_7x7.grid.binned == None

    def test__inversion_pixel_limit(self, sub_mask_7x7):
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, inversion_pixel_limit=2
        )

        assert lens_imaging_data_7x7.inversion_pixel_limit == 2

        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, inversion_pixel_limit=5
        )

        assert lens_imaging_data_7x7.inversion_pixel_limit == 5

    def test__hyper_noise_map_max(self, sub_mask_7x7):
        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, hyper_noise_map_max=10.0
        )

        assert lens_imaging_data_7x7.hyper_noise_map_max == 10.0

        lens_imaging_data_7x7 = al.AbstractLensData(
            mask=sub_mask_7x7, hyper_noise_map_max=20.0
        )

        assert lens_imaging_data_7x7.hyper_noise_map_max == 20.0


class TestLensImagingData(object):
    def test__attributes(self, imaging_data_7x7, lens_imaging_data_7x7):
        assert lens_imaging_data_7x7.pixel_scale == imaging_data_7x7.pixel_scale
        assert lens_imaging_data_7x7.pixel_scale == 1.0

        assert (
            lens_imaging_data_7x7.image(return_in_2d=True, return_masked=False)
            == imaging_data_7x7.image
        ).all()
        assert (
            lens_imaging_data_7x7.image(return_in_2d=True, return_masked=False)
            == np.ones((7, 7))
        ).all()

        assert (
            lens_imaging_data_7x7.noise_map(return_in_2d=True, return_masked=False)
            == imaging_data_7x7.noise_map
        ).all()
        assert (
            lens_imaging_data_7x7.noise_map(return_in_2d=True, return_masked=False)
            == 2.0 * np.ones((7, 7))
        ).all()

        assert (lens_imaging_data_7x7.psf == imaging_data_7x7.psf).all()
        assert (lens_imaging_data_7x7.psf == np.ones((3, 3))).all()

        assert lens_imaging_data_7x7.trimmed_psf_shape == (3, 3)

    def test__masking(self, lens_imaging_data_7x7):

        assert (
            lens_imaging_data_7x7._mask_1d == np.full(fill_value=False, shape=(9))
        ).all()
        assert (lens_imaging_data_7x7._image_1d == np.ones(9)).all()
        assert (lens_imaging_data_7x7._noise_map_1d == 2.0 * np.ones(9)).all()

        assert (
            lens_imaging_data_7x7.mask
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            lens_imaging_data_7x7.image(return_in_2d=True, return_masked=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            lens_imaging_data_7x7.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

    def test__grids(
        self, lens_imaging_data_7x7, grid_7x7, sub_grid_7x7, blurring_grid_7x7
    ):

        assert (lens_imaging_data_7x7.grid.unlensed_unsubbed_1d == grid_7x7).all()
        assert (lens_imaging_data_7x7.grid == sub_grid_7x7).all()
        assert (lens_imaging_data_7x7.blurring_grid == blurring_grid_7x7).all()

    def test__pixel_scale_interpolation_grid_input__grids_include_interpolator_on_blurring_grid(
        self, imaging_data_7x7, sub_mask_7x7
    ):

        lens_imaging_data_7x7 = al.LensImagingData(
            imaging_data=imaging_data_7x7,
            mask=sub_mask_7x7,
            pixel_scale_interpolation_grid=1.0,
        )

        grid = al.Grid.from_mask(mask=sub_mask_7x7)
        new_grid = grid.new_grid_with_interpolator(pixel_scale_interpolation_grid=1.0)

        blurring_grid = al.Grid.blurring_grid_from_mask_and_psf_shape(
            mask=sub_mask_7x7, psf_shape=(3, 3)
        )
        new_blurring_grid = blurring_grid.new_grid_with_interpolator(
            pixel_scale_interpolation_grid=1.0
        )

        assert (lens_imaging_data_7x7.grid == new_grid).all()
        assert (
            lens_imaging_data_7x7.grid.interpolator.vtx == new_grid.interpolator.vtx
        ).all()
        assert (
            lens_imaging_data_7x7.grid.interpolator.wts == new_grid.interpolator.wts
        ).all()

        assert (lens_imaging_data_7x7.blurring_grid == new_blurring_grid).all()
        assert (
            lens_imaging_data_7x7.blurring_grid.interpolator.vtx
            == new_blurring_grid.interpolator.vtx
        ).all()
        assert (
            lens_imaging_data_7x7.blurring_grid.interpolator.wts
            == new_blurring_grid.interpolator.wts
        ).all()

    def test__convolvers(self, lens_imaging_data_7x7):
        assert type(lens_imaging_data_7x7.convolver) == al.Convolver

    def test__different_imaging_data_without_mock_objects__customize_constructor_inputs(
        self
    ):

        psf = al.PSF(np.ones((7, 7)), 1)
        imaging_data = al.ImagingData(
            np.ones((19, 19)),
            pixel_scale=3.0,
            psf=psf,
            noise_map=2.0 * np.ones((19, 19)),
        )
        mask = al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
            shape=(19, 19), pixel_scale=1.0, invert=True, sub_size=8
        )
        mask[9, 9] = False

        lens_imaging_data_7x7 = al.LensImagingData(
            imaging_data=imaging_data,
            mask=mask,
            trimmed_psf_shape=(7, 7),
            positions=[np.array([[1.0, 1.0]])],
            positions_threshold=1.0,
        )

        assert (
            lens_imaging_data_7x7.image(return_in_2d=True, return_masked=False)
            == np.ones((19, 19))
        ).all()
        assert (
            lens_imaging_data_7x7.noise_map(return_in_2d=True, return_masked=False)
            == 2.0 * np.ones((19, 19))
        ).all()
        assert (lens_imaging_data_7x7.psf == np.ones((7, 7))).all()

        assert lens_imaging_data_7x7.sub_size == 8
        assert lens_imaging_data_7x7.convolver.psf.shape == (7, 7)
        assert (lens_imaging_data_7x7.positions[0] == np.array([[1.0, 1.0]])).all()
        assert lens_imaging_data_7x7.positions_threshold == 1.0

        assert lens_imaging_data_7x7.trimmed_psf_shape == (7, 7)

    def test__lens_imaging_data_7x7_with_modified_image(self, lens_imaging_data_7x7):

        lens_imaging_data_7x7 = lens_imaging_data_7x7.new_lens_imaging_data_with_modified_image(
            modified_image=8.0 * np.ones((7, 7))
        )

        assert (
            lens_imaging_data_7x7.image(return_in_2d=True, return_masked=False)
            == 8.0 * np.ones((7, 7))
        ).all()

        assert (lens_imaging_data_7x7._image_1d == 8.0 * np.ones(9)).all()

        assert (
            lens_imaging_data_7x7.image(return_in_2d=True, return_masked=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 8.0, 8.0, 8.0, 0.0, 0.0],
                    [0.0, 0.0, 8.0, 8.0, 8.0, 0.0, 0.0],
                    [0.0, 0.0, 8.0, 8.0, 8.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

    def test__lens_imaging_data_6x6_with_binned_up_imaging_data(
        self, lens_imaging_data_6x6
    ):
        binned_up_psf = lens_imaging_data_6x6.imaging_data.psf.new_psf_with_rescaled_odd_dimensioned_array(
            rescale_factor=0.5
        )

        lens_imaging_data_6x6 = lens_imaging_data_6x6.new_lens_imaging_data_with_binned_up_imaging_data_and_mask(
            bin_up_factor=2
        )

        assert (
            lens_imaging_data_6x6.image(return_in_2d=True, return_masked=False)
            == np.ones((3, 3))
        ).all()
        assert (lens_imaging_data_6x6.psf == binned_up_psf).all()
        assert (
            lens_imaging_data_6x6.noise_map(return_in_2d=True, return_masked=False)
            == np.ones((3, 3))
        ).all()
        assert (
            lens_imaging_data_6x6.imaging_data.background_noise_map
            == 1.5 * np.ones((3, 3))
        ).all()
        assert (
            lens_imaging_data_6x6.imaging_data.poisson_noise_map
            == 2.0 * np.ones((3, 3))
        ).all()
        assert (
            lens_imaging_data_6x6.imaging_data.exposure_time_map
            == 20.0 * np.ones((3, 3))
        ).all()
        assert (
            lens_imaging_data_6x6.imaging_data.background_sky_map
            == 6.0 * np.ones((3, 3))
        ).all()

        assert (
            lens_imaging_data_6x6.mask
            == np.array([[True, True, True], [True, False, True], [True, True, True]])
        ).all()

        assert (lens_imaging_data_6x6._image_1d == np.ones((1))).all()
        assert (lens_imaging_data_6x6._noise_map_1d == np.ones((1))).all()

    def test__lens_imaging_data_7x7_with_signal_to_noise_limit(
        self, imaging_data_7x7, lens_imaging_data_7x7
    ):

        lens_data_snr_limit = lens_imaging_data_7x7.new_lens_imaging_data_with_signal_to_noise_limit(
            signal_to_noise_limit=0.25
        )

        assert lens_data_snr_limit.pixel_scale == imaging_data_7x7.pixel_scale
        assert lens_data_snr_limit.pixel_scale == 1.0

        assert (
            lens_data_snr_limit.image(return_in_2d=True, return_masked=False)
            == imaging_data_7x7.image
        ).all()
        assert (
            lens_data_snr_limit.image(return_in_2d=True, return_masked=False)
            == np.ones((7, 7))
        ).all()

        assert (
            lens_data_snr_limit.noise_map(return_in_2d=True, return_masked=False)
            == 4.0 * np.ones((7, 7))
        ).all()

        assert (lens_data_snr_limit.psf == imaging_data_7x7.psf).all()
        assert (lens_data_snr_limit.psf == np.ones((3, 3))).all()

        assert lens_data_snr_limit.trimmed_psf_shape == (3, 3)

        assert (lens_data_snr_limit._image_1d == np.ones(9)).all()
        assert (lens_data_snr_limit._noise_map_1d == 4.0 * np.ones(9)).all()

        assert (
            lens_data_snr_limit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()


class TestLensUVPlaneData(object):
    def test__attributes(self, uv_plane_data_7, lens_uv_plane_data_7, sub_mask_7x7):

        assert lens_uv_plane_data_7.pixel_scale == uv_plane_data_7.pixel_scale
        assert lens_uv_plane_data_7.pixel_scale == 1.0

        assert (
            lens_uv_plane_data_7.visibilities() == uv_plane_data_7.visibilities
        ).all()
        assert (lens_uv_plane_data_7.visibilities() == np.ones((7, 2))).all()

        assert (lens_uv_plane_data_7.noise_map() == uv_plane_data_7.noise_map).all()
        assert (lens_uv_plane_data_7.noise_map() == 2.0 * np.ones((7))).all()
        assert (
            lens_uv_plane_data_7.noise_map(return_x2=True)[:, 0] == 2.0 * np.ones((7))
        ).all()
        assert (
            lens_uv_plane_data_7.noise_map(return_x2=True)[:, 1] == 2.0 * np.ones((7))
        ).all()

        assert (
            lens_uv_plane_data_7.visibilities_mask
            == np.full(fill_value=False, shape=(7, 2))
        ).all()

        assert (lens_uv_plane_data_7.primary_beam == uv_plane_data_7.primary_beam).all()
        assert (lens_uv_plane_data_7.primary_beam == np.ones((3, 3))).all()
        assert lens_uv_plane_data_7.trimmed_primary_beam_shape == (3, 3)

        assert (
            lens_uv_plane_data_7.uv_plane_data.uv_wavelengths
            == uv_plane_data_7.uv_wavelengths
        ).all()
        assert lens_uv_plane_data_7.uv_plane_data.uv_wavelengths[0, 0] == -55636.4609375

    def test__grids(self, lens_uv_plane_data_7, grid_7x7, sub_grid_7x7):
        assert (lens_uv_plane_data_7.grid.unlensed_unsubbed_1d == grid_7x7).all()
        assert (lens_uv_plane_data_7.grid == sub_grid_7x7).all()

    def test__transformer(self, lens_uv_plane_data_7):
        assert type(lens_uv_plane_data_7.transformer) == al.Transformer

    def test__different_uv_plane_data_without_mock_objects__customize_constructor_inputs(
        self
    ):
        primary_beam = al.PrimaryBeam(np.ones((7, 7)), 1)
        uv_plane_data = al.UVPlaneData(
            shape=(2, 2),
            visibilities=np.ones((19, 2)),
            pixel_scale=3.0,
            primary_beam=primary_beam,
            noise_map=2.0 * np.ones((19,)),
            uv_wavelengths=3.0 * np.ones((19, 2)),
        )
        mask = al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
            shape=(19, 19), pixel_scale=1.0, invert=True, sub_size=8
        )
        mask[9, 9] = False

        lens_uv_plane_data_7 = al.LensUVPlaneData(
            uv_plane_data=uv_plane_data,
            mask=mask,
            trimmed_primary_beam_shape=(7, 7),
            positions=[np.array([[1.0, 1.0]])],
            positions_threshold=1.0,
        )

        assert (lens_uv_plane_data_7.visibilities() == np.ones((19, 2))).all()
        assert (lens_uv_plane_data_7.noise_map() == 2.0 * np.ones((19,))).all()
        assert (
            lens_uv_plane_data_7.uv_plane_data.uv_wavelengths == 3.0 * np.ones((19, 2))
        ).all()
        assert (lens_uv_plane_data_7.primary_beam == np.ones((7, 7))).all()

        assert lens_uv_plane_data_7.sub_size == 8
        assert (lens_uv_plane_data_7.positions[0] == np.array([[1.0, 1.0]])).all()
        assert lens_uv_plane_data_7.positions_threshold == 1.0

        assert lens_uv_plane_data_7.trimmed_primary_beam_shape == (7, 7)

    def test__lens_uv_plane_data_7_with_modified_visibilities(
        self, lens_uv_plane_data_7
    ):
        lens_uv_plane_data_7 = lens_uv_plane_data_7.new_lens_imaging_data_with_modified_visibilities(
            modified_visibilities=8.0 * np.ones((7, 2))
        )

        assert (lens_uv_plane_data_7.visibilities() == 8.0 * np.ones((7, 2))).all()



from autolens.array import mask as msk
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.lens.lens_fit import lens_imaging_fit
from autolens.lens import lens_data as ld
from autolens.lens.plotters import lens_imaging_fit_plotters
from test.simulation import simulation_util

# In this tutorial, we'll introduce a new pixelization, called an adaptive-pixelization. This pixelization doesn't use
# uniform grid of rectangular pixels, but instead uses ir'Voronoi' pixels. So, why would we want to do that?
# Lets take another look at the rectangular grid, and think about its weakness.

# Lets quickly remind ourselves of the image, and the 3.0" circular mask we'll use to mask it.
imaging_data = simulation_util.load_test_imaging_data(
    data_type="lens_light_dev_vaucouleurs", data_resolution="LSST"
)
mask = al.Mask.elliptical(
    shape=imaging_data.shape,
    pixel_scale=imaging_data.pixel_scale,
    major_axis_radius_arcsec=3.0,
    axis_ratio=1.0,
    phi=0.0,
    centre=(0.0, 0.0),
)

# imaging_plotters.plot_imaging_subplot(imaging_data=imaging_data, mask=mask, zoom_around_mask=True, aspect='equal')
# imaging_plotters.plot_imaging_subplot(imaging_data=imaging_data, mask=mask, zoom_around_mask=True, aspect='auto')

# imaging_plotters.plot_image(imaging_data=imaging_data, mask=mask, zoom_around_mask=True, aspect='square')
# imaging_plotters.plot_image(imaging_data=imaging_data, mask=mask, zoom_around_mask=True, aspect='equal')

# The lines of code below do everything we're used to, that is, setup an image and its al.ogrid, mask it, trace it
# via a tracer, setup the rectangular mapper, etc.
lens_galaxy = al.Galaxy(
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(1.0, 1.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    )
)
source_galaxy = al.Galaxy(
    pixelization=al.pixelizations.VoronoiMagnification(shape=(20, 20)),
    regularization=al.regularization.Constant(coefficient=1.0),
)

lens_data = al.LensData(imaging_data=imaging_data, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])
fit = al.LensImageFit.from_lens_data_and_tracer(lens_data=lens_data, tracer=tracer)


lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    should_plot_image_plane_pix=True,
    aspect="auto",
)

lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    should_plot_image_plane_pix=True,
    aspect="equal",
)

lens_imaging_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    should_plot_image_plane_pix=True,
    aspect="square",
)



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens.plotters import plotter_util, array_plotters
from autolens.model.inversion.plotters import mapper_plotters
from autolens.model.inversion import mappers


def plot_inversion_subplot(
    inversion,
    mask=None,
    positions=None,
    grid=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    output_path=None,
    output_format="show",
    output_filename="inversion_subplot",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=6
    )

    if figsize is None:
        figsize = figsize_tool

    ratio = float(
        (
            inversion.mapper.geometry.arc_second_maxima[1]
            - inversion.mapper.geometry.arc_second_minima[1]
        )
        / (
            inversion.mapper.geometry.arc_second_maxima[0]
            - inversion.mapper.geometry.arc_second_minima[0]
        )
    )

    if aspect is "square":
        aspect_inv = ratio
    elif aspect is "auto":
        aspect_inv = 1.0 / ratio
    elif aspect is "equal":
        aspect_inv = 1.0

    plt.figure(figsize=figsize)

    plt.subplot(rows, columns, 1)

    plot_reconstructed_image(
        inversion=inversion,
        mask=mask,
        positions=positions,
        grid=grid,
        as_subplot=True,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plt.subplot(rows, columns, 2, aspect=float(aspect_inv))

    plot_pixelization_values(
        inversion=inversion,
        positions=None,
        should_plot_grid=False,
        should_plot_centres=False,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=None,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename=None,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 3, aspect=float(aspect_inv))

    plot_pixelization_residual_map(
        inversion=inversion,
        positions=None,
        should_plot_grid=False,
        should_plot_centres=False,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=None,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename=None,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 4, aspect=float(aspect_inv))

    plot_pixelization_normalized_residual_map(
        inversion=inversion,
        positions=None,
        should_plot_grid=False,
        should_plot_centres=False,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=None,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename=None,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 5, aspect=float(aspect_inv))

    plot_pixelization_chi_squared_map(
        inversion=inversion,
        positions=None,
        should_plot_grid=False,
        should_plot_centres=False,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=None,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename=None,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 6, aspect=float(aspect_inv))

    plot_pixelization_regularization_weights(
        inversion=inversion,
        positions=None,
        should_plot_grid=False,
        should_plot_centres=False,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=None,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename=None,
        output_format=output_format,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def plot_reconstructed_image(
    inversion,
    mask=None,
    positions=None,
    grid=None,
    as_subplot=False,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="reconstructed_inversion_image",
):

    array_plotters.plot_array(
        array=inversion.reconstructed_data_2d,
        mask=mask,
        positions=positions,
        grid=grid,
        as_subplot=as_subplot,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_pixelization_values(
    inversion,
    plot_origin=True,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Pixelization",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="inversion_pixelization_values",
):

    if output_format is "fits":
        return

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    plot_inversion_with_source_values(
        inversion=inversion,
        source_pixel_values=inversion.pixelization_values,
        plot_origin=plot_origin,
        positions=positions,
        should_plot_centres=should_plot_centres,
        should_plot_grid=should_plot_grid,
        should_plot_border=should_plot_border,
        image_pixels=image_pixels,
        source_pixels=source_pixels,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plotter_util.close_figure(as_subplot=as_subplot)


def plot_pixelization_residual_map(
    inversion,
    plot_origin=True,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Pixelization Residual-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="inversion_pixelization_residual_map",
):

    if output_format is "fits":
        return

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    plot_inversion_with_source_values(
        inversion=inversion,
        source_pixel_values=inversion.pixelization_residual_map,
        plot_origin=plot_origin,
        positions=positions,
        should_plot_centres=should_plot_centres,
        should_plot_grid=should_plot_grid,
        should_plot_border=should_plot_border,
        image_pixels=image_pixels,
        source_pixels=source_pixels,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plotter_util.close_figure(as_subplot=as_subplot)


def plot_pixelization_normalized_residual_map(
    inversion,
    plot_origin=True,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Pixelization Normalized Residual Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="inversion_pixelization_normalized_residual_map",
):

    if output_format is "fits":
        return

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    plot_inversion_with_source_values(
        inversion=inversion,
        source_pixel_values=inversion.pixelization_normalized_residual_map,
        plot_origin=plot_origin,
        positions=positions,
        should_plot_centres=should_plot_centres,
        should_plot_grid=should_plot_grid,
        should_plot_border=should_plot_border,
        image_pixels=image_pixels,
        source_pixels=source_pixels,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plotter_util.close_figure(as_subplot=as_subplot)


def plot_pixelization_chi_squared_map(
    inversion,
    plot_origin=True,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Pixelization Chi-Squared Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="inversion_pixelization_chi_squared_map",
):

    if output_format is "fits":
        return

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    plot_inversion_with_source_values(
        inversion=inversion,
        source_pixel_values=inversion.pixelization_chi_squared_map,
        plot_origin=plot_origin,
        positions=positions,
        should_plot_centres=should_plot_centres,
        should_plot_grid=should_plot_grid,
        should_plot_border=should_plot_border,
        image_pixels=image_pixels,
        source_pixels=source_pixels,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plotter_util.close_figure(as_subplot=as_subplot)


def plot_pixelization_regularization_weights(
    inversion,
    plot_origin=True,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Pixelization Regularization Weights",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="inversion_pixelization_regularization_weights",
):

    if output_format is "fits":
        return

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    regularization_weights = inversion.regularization.regularization_weights_from_mapper(
        mapper=inversion.mapper
    )

    plot_inversion_with_source_values(
        inversion=inversion,
        source_pixel_values=regularization_weights,
        plot_origin=plot_origin,
        positions=positions,
        should_plot_centres=should_plot_centres,
        should_plot_grid=should_plot_grid,
        should_plot_border=should_plot_border,
        image_pixels=image_pixels,
        source_pixels=source_pixels,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plotter_util.close_figure(as_subplot=as_subplot)


def plot_inversion_with_source_values(
    inversion,
    source_pixel_values,
    plot_origin=True,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Reconstructed Pixelization",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="pixelization_source_values",
):

    if isinstance(inversion.mapper, mappers.RectangularMapper):

        reconstructed_pixelization = inversion.mapper.reconstructed_pixelization_from_solution_vector(
            solution_vector=source_pixel_values
        )

        origin = get_origin(image=reconstructed_pixelization, plot_origin=plot_origin)

        array_plotters.plot_array(
            array=reconstructed_pixelization,
            origin=origin,
            positions=positions,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title=title,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_filename=output_filename,
        )

        mapper_plotters.plot_rectangular_mapper(
            mapper=inversion.mapper,
            should_plot_centres=should_plot_centres,
            should_plot_grid=should_plot_grid,
            should_plot_border=should_plot_border,
            image_pixels=image_pixels,
            source_pixels=source_pixels,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            title=title,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
        )

        plotter_util.output_figure(
            array=reconstructed_pixelization,
            as_subplot=as_subplot,
            output_path=output_path,
            output_filename=output_filename,
            output_format=output_format,
        )

    elif isinstance(inversion.mapper, mappers.VoronoiMapper):

        mapper_plotters.plot_voronoi_mapper(
            mapper=inversion.mapper,
            source_pixel_values=source_pixel_values,
            should_plot_centres=should_plot_centres,
            should_plot_grid=should_plot_grid,
            should_plot_border=should_plot_border,
            image_pixels=image_pixels,
            source_pixels=source_pixels,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            title=title,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
        )

        plotter_util.output_figure(
            array=None,
            as_subplot=as_subplot,
            output_path=output_path,
            output_filename=output_filename,
            output_format=output_format,
        )


def get_origin(image, plot_origin):

    if plot_origin:
        return image.origin
    else:
        return None



import os
import numpy as np

import autofit as af
import autolens as al
from test.integration import integration_util

test_type = "galaxy_fit"
test_name = "potential"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"


def galaxy_fit_phase():

    pixel_scale = 0.1
    image_shape = (150, 150)

    integration_util.reset_paths(test_name=test_name, output_path=output_path)

    grid = al.Grid.from_shape_pixel_scale_and_sub_size(
        shape=image_shape, pixel_scale=pixel_scale, sub_size=4
    )

    galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        ),
    )

    potential = galaxy.convergence_from_grid(
        galaxies=[galaxy], grid=grid, return_in_2d=True, bypass_decorator=False
    )

    noise_map = al.ScaledSquarePixelArray(
        array=np.ones(potential.shape), pixel_scale=pixel_scale
    )

    data = al.GalaxyData(image=potential, noise_map=noise_map, pixel_scale=pixel_scale)

    phase1 = al.PhaseGalaxy(
        phase_name=test_name + "/",
        galaxies=dict(
            gal=al.GalaxyModel(redshift=0.5, light=al.mass_profiles.SphericalIsothermal)
        ),
        use_convergence=True,
        sub_size=4,
        optimizer_class=af.MultiNest,
    )

    phase1.run(galaxy_data=[data])


if __name__ == "__main__":
    galaxy_fit_phase()



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt
import numpy as np
import itertools

from autolens import exc
from autolens.plotters import plotter_util


def plot_grid(
    grid,
    colors=None,
    axis_limits=None,
    points=None,
    lines=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(12, 8),
    pointsize=5,
    pointcolor="k",
    xyticksize=16,
    cmap="jet",
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Grid",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    symmetric_around_centre=True,
    output_path=None,
    output_format="show",
    output_filename="grid",
):
    """Plot a grid of (y,x) Cartesian coordinates as a scatter plot of points.

    Parameters
    -----------
    grid : data_type.array.grids.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2).
    axis_limits : []
        The axis limits of the figure on which the grid is plotted, following [xmin, xmax, ymin, ymax].
    points : []
        A set of points that are plotted in a different colour for emphasis (e.g. to show the mappings between \
        different planes).
    as_subplot : bool
        Whether the grid is plotted as part of a subplot, in which case the grid figure is not opened / closed.
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    figsize : (int, int)
        The size of the figure in (rows, columns).
    pointsize : int
        The size of the points plotted on the grid.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    title : str
        The text of the title.
    titlesize : int
        The size of of the title of the figure.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    output_path : str
        The path on the hard-disk where the figure is output.
    output_filename : str
        The filename of the figure that is output.
    output_format : str
        The format the figue is output:
        'show' - display on computer screen.
        'png' - output to hard-disk as a png.
    """

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)
    grid = convert_grid_units(
        grid_arcsec=grid, units=units, kpc_per_arcsec=kpc_per_arcsec
    )

    if colors is not None:

        plt.cm.get_cmap(cmap)

    plt.scatter(
        y=np.asarray(grid[:, 0]),
        x=np.asarray(grid[:, 1]),
        c=colors,
        s=pointsize,
        marker=".",
        cmap=cmap,
    )

    if colors is not None:

        plotter_util.set_colorbar(
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
        )

    plotter_util.set_title(title=title, titlesize=titlesize)
    set_xy_labels(
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
    )

    set_axis_limits(
        axis_limits=axis_limits,
        grid=grid,
        symmetric_around_centre=symmetric_around_centre,
    )
    plot_points(grid=grid, points=points, pointcolor=pointcolor)
    plotter_util.plot_lines(line_lists=lines)

    plt.tick_params(labelsize=xyticksize)
    plotter_util.output_figure(
        array=None,
        as_subplot=as_subplot,
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plotter_util.close_figure(as_subplot=as_subplot)


def convert_grid_units(grid_arcsec, units, kpc_per_arcsec):
    """Convert the grid from its input units (arc-seconds) to the input unit (e.g. retain arc-seconds) or convert to \
    another set of units (kiloparsecs).

    Parameters
    -----------
    grid_arcsec : ndarray or data_type.array.grids.Grid
        The (y,x) coordinates of the grid in arc-seconds, in an array of shape (total_coordinates, 2).
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    """

    if units in "arcsec" or kpc_per_arcsec is None:
        return grid_arcsec
    elif units in "kpc":
        return grid_arcsec * kpc_per_arcsec


def set_xy_labels(units, kpc_per_arcsec, xlabelsize, ylabelsize, xyticksize):
    """Set the x and y labels of the figure, and set the fontsize of those labels.

    The x and y labels are always the distance scales, thus the labels are either arc-seconds or kpc and depend on the \
    units the figure is plotted in.

    Parameters
    -----------
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    """
    if units in "arcsec" or kpc_per_arcsec is None:

        plt.xlabel("x (arcsec)", fontsize=xlabelsize)
        plt.ylabel("y (arcsec)", fontsize=ylabelsize)

    elif units in "kpc":

        plt.xlabel("x (kpc)", fontsize=xlabelsize)
        plt.ylabel("y (kpc)", fontsize=ylabelsize)

    else:
        raise exc.PlottingException(
            "The units supplied to the plotted are not a valid string (must be pixels | "
            "arcsec | kpc)"
        )

    plt.tick_params(labelsize=xyticksize)


def set_axis_limits(axis_limits, grid, symmetric_around_centre):
    """Set the axis limits of the figure the grid is plotted on.

    Parameters
    -----------
    axis_limits : []
        The axis limits of the figure on which the grid is plotted, following [xmin, xmax, ymin, ymax].
    """
    if axis_limits is not None:
        plt.axis(axis_limits)
    elif symmetric_around_centre:
        ymin = np.min(grid[:, 0])
        ymax = np.max(grid[:, 0])
        xmin = np.min(grid[:, 1])
        xmax = np.max(grid[:, 1])
        x = np.max([np.abs(xmin), np.abs(xmax)])
        y = np.max([np.abs(ymin), np.abs(ymax)])
        axis_limits = [-x, x, -y, y]
        plt.axis(axis_limits)


def plot_points(grid, points, pointcolor):
    """Plot a subset of points in a different color, to highlight a specifc region of the grid (e.g. how certain \
    pixels map between different planes).

    Parameters
    -----------
    grid : ndarray or data_type.array.grids.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2).
    points : []
        A set of points that are plotted in a different colour for emphasis (e.g. to show the mappings between \
        different planes).
    pointcolor : str or None
        The color the points should be plotted. If None, the points are iterated through a cycle of colors.
    """
    if points is not None:

        if pointcolor is None:

            point_colors = itertools.cycle(["y", "r", "k", "g", "m"])
            for point_set in points:
                plt.scatter(
                    y=np.asarray(grid[point_set, 0]),
                    x=np.asarray(grid[point_set, 1]),
                    s=8,
                    color=next(point_colors),
                )

        else:

            for point_set in points:
                plt.scatter(
                    y=np.asarray(grid[point_set, 0]),
                    x=np.asarray(grid[point_set, 1]),
                    s=8,
                    color=pointcolor,
                )



import autolens as al
import pytest
import os


@pytest.fixture(name="lens_fit_plotter_path")
def make_lens_fit_plotter_setup():
    return "{}/../../test_files/plotting/fit/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__fit_sub_plot(
    lens_imaging_fit_x2_plane_7x7, lens_fit_plotter_path, plot_patch
):

    al.lens_imaging_fit_plotters.plot_fit_subplot(
        fit=lens_imaging_fit_x2_plane_7x7,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "lens_fit.png" in plot_patch.paths


def test__fit_for_plane_subplot(
    lens_imaging_fit_x1_plane_7x7,
    lens_imaging_fit_x2_plane_7x7,
    lens_fit_plotter_path,
    plot_patch,
):

    al.lens_imaging_fit_plotters.plot_fit_subplot_for_plane(
        fit=lens_imaging_fit_x1_plane_7x7,
        plane_index=0,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "lens_fit_plane_0.png" in plot_patch.paths

    al.lens_imaging_fit_plotters.plot_fit_subplot_for_plane(
        fit=lens_imaging_fit_x2_plane_7x7,
        plane_index=0,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "lens_fit_plane_0.png" in plot_patch.paths

    al.lens_imaging_fit_plotters.plot_fit_subplot_for_plane(
        fit=lens_imaging_fit_x2_plane_7x7,
        plane_index=1,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "lens_fit_plane_1.png" in plot_patch.paths


def test__fit_for_planes_subplot(
    lens_imaging_fit_x1_plane_7x7,
    lens_imaging_fit_x2_plane_7x7,
    lens_fit_plotter_path,
    plot_patch,
):

    al.lens_imaging_fit_plotters.plot_fit_subplot_of_planes(
        fit=lens_imaging_fit_x1_plane_7x7,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "lens_fit_plane_0.png" in plot_patch.paths

    al.lens_imaging_fit_plotters.plot_fit_subplot_of_planes(
        fit=lens_imaging_fit_x2_plane_7x7,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "lens_fit_plane_0.png" in plot_patch.paths
    assert lens_fit_plotter_path + "lens_fit_plane_1.png" in plot_patch.paths


def test__fit_individuals__source_and_lens__depedent_on_input(
    lens_imaging_fit_x1_plane_7x7,
    lens_imaging_fit_x2_plane_7x7,
    lens_fit_plotter_path,
    plot_patch,
):

    al.lens_imaging_fit_plotters.plot_fit_individuals(
        fit=lens_imaging_fit_x1_plane_7x7,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_chi_squared_map=True,
        should_plot_subtracted_images_of_planes=True,
        should_plot_model_images_of_planes=True,
        should_plot_plane_images_of_planes=True,
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "fit_image.png" in plot_patch.paths

    assert lens_fit_plotter_path + "fit_noise_map.png" not in plot_patch.paths

    assert lens_fit_plotter_path + "fit_signal_to_noise_map.png" not in plot_patch.paths

    assert lens_fit_plotter_path + "fit_model_image.png" in plot_patch.paths

    assert lens_fit_plotter_path + "fit_residual_map.png" not in plot_patch.paths

    assert (
        lens_fit_plotter_path + "fit_normalized_residual_map.png"
        not in plot_patch.paths
    )

    assert lens_fit_plotter_path + "fit_chi_squared_map.png" in plot_patch.paths

    assert (
        lens_fit_plotter_path + "fit_subtracted_image_of_plane_0.png"
        in plot_patch.paths
    )

    assert lens_fit_plotter_path + "fit_model_image_of_plane_0.png" in plot_patch.paths

    assert lens_fit_plotter_path + "fit_plane_image_of_plane_0.png" in plot_patch.paths

    al.lens_imaging_fit_plotters.plot_fit_individuals(
        fit=lens_imaging_fit_x2_plane_7x7,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_chi_squared_map=True,
        should_plot_subtracted_images_of_planes=True,
        should_plot_model_images_of_planes=True,
        should_plot_plane_images_of_planes=True,
        output_path=lens_fit_plotter_path,
        output_format="png",
    )

    assert lens_fit_plotter_path + "fit_image.png" in plot_patch.paths

    assert lens_fit_plotter_path + "fit_noise_map.png" not in plot_patch.paths

    assert lens_fit_plotter_path + "fit_signal_to_noise_map.png" not in plot_patch.paths

    assert lens_fit_plotter_path + "fit_model_image.png" in plot_patch.paths

    assert lens_fit_plotter_path + "fit_residual_map.png" not in plot_patch.paths

    assert (
        lens_fit_plotter_path + "fit_normalized_residual_map.png"
        not in plot_patch.paths
    )

    assert lens_fit_plotter_path + "fit_chi_squared_map.png" in plot_patch.paths

    assert (
        lens_fit_plotter_path + "fit_subtracted_image_of_plane_0.png"
        in plot_patch.paths
    )
    assert (
        lens_fit_plotter_path + "fit_subtracted_image_of_plane_1.png"
        in plot_patch.paths
    )

    assert lens_fit_plotter_path + "fit_model_image_of_plane_0.png" in plot_patch.paths
    assert lens_fit_plotter_path + "fit_model_image_of_plane_1.png" in plot_patch.paths

    assert lens_fit_plotter_path + "fit_plane_image_of_plane_0.png" in plot_patch.paths
    assert lens_fit_plotter_path + "fit_plane_image_of_plane_1.png" in plot_patch.paths
